@@ -2,7 +2,8 @@ use std::{collections::BTreeMap, iter::repeat};
 
 use criterion::{black_box, criterion_group, criterion_main, Criterion};
 use evalexpr::{build_operator_tree, ContextWithMutableVariables, HashMapContext, Node, Value};
-use exmex::{parse_with_default_ops, BinOp, FlatEx, Operator, OwnedFlatEx};
+use exmex::OwnedFlatEx;
+use exmex::{ops_factory, prelude::*, BinOp, Operator};
 use fasteval::{Compiler, Evaler, Instruction, Slab};
 use itertools::{izip, Itertools};
 
@@ -11,6 +12,9 @@ use rsc::{
     lexer::tokenize,
     parser::{parse, Expr},
 };
+#[cfg(feature = "serde")]
+use serde::{Deserialize, Serialize};
+
 const N: usize = 4;
 
 const BENCH_EXPRESSIONS_NAMES: [&str; N] = ["sin", "power", "nested", "compile"];
@@ -36,6 +40,15 @@ const BENCH_X_RANGE: (usize, usize) = (0, 5);
 const BENCH_Y: f64 = 3.0;
 const BENCH_Z: f64 = 4.0;
 
+const BENCH_PARSEVAL_STRS: [&str; N] = [
+    "2*3^2",
+    "sin(-(sin(2)))*2",
+    "-1*(1.3+(-0.7)*(2-1/10))",
+    "log(log2(2))*tan(2)+exp(1.5)",
+];
+
+const BENCH_PARSEVAL_REFS: [f64; N] = [18.0, -1.5781446871457767, 0.03, 4.4816890703380645];
+
 fn bench_ref_values() -> Vec<Vec<f64>> {
     BENCH_EXPRESSIONS_REFS
         .iter()
@@ -67,6 +80,17 @@ fn run_benchmark<F: FnMut(f64) -> f64>(funcs: Vec<F>, eval_name: &str, c: &mut C
     }
 }
 
+fn run_benchmark_parseval(func: fn(&str) -> f64, eval_name: &str, c: &mut Criterion) {
+    c.bench_function(format!("parseval {}", eval_name).as_str(), |b| {
+        b.iter(|| {
+            for (exp_str, ref_val) in izip!(BENCH_PARSEVAL_STRS.iter(), BENCH_PARSEVAL_REFS.iter())
+            {
+                assert_float_eq(func(black_box(exp_str)), *ref_val);
+            }
+        })
+    });
+}
+
 fn run_benchmark_parse<'a, T, F: Fn(&'a [&str]) -> Vec<T>>(
     func: F,
     parse_name: &str,
@@ -82,7 +106,7 @@ fn run_benchmark_parse<'a, T, F: Fn(&'a [&str]) -> Vec<T>>(
 fn exmex_parse_owned<'a>(strings: &'a [&str]) -> Vec<OwnedFlatEx<f64>> {
     strings
         .iter()
-        .map(|expr_str| OwnedFlatEx::from_flatex(parse_with_default_ops::<f64>(expr_str).unwrap()))
+        .map(|expr_str| OwnedFlatEx::<f64>::from_str(expr_str).unwrap())
         .collect::<Vec<_>>()
 }
 
@@ -93,65 +117,67 @@ fn exmex_bench_parse_owned(c: &mut Criterion) {
 fn exmex_parse<'a>(strings: &'a [&str]) -> Vec<FlatEx<'a, f64>> {
     strings
         .iter()
-        .map(|expr_str| parse_with_default_ops::<f64>(expr_str).unwrap())
+        .map(|expr_str| FlatEx::<f64>::from_str(expr_str).unwrap())
         .collect::<Vec<_>>()
 }
 
 fn exmex_bench_parse(c: &mut Criterion) {
     run_benchmark_parse(exmex_parse, "exmex_parse", c);
 }
+ops_factory!(
+    OnlyNeededOperators,
+    f64,
+    Operator {
+        repr: "^",
+        bin_op: Some(BinOp {
+            apply: |a: f64, b| a.powf(b),
+            prio: 2,
+        }),
+        unary_op: None,
+    },
+    Operator {
+        repr: "*",
+        bin_op: Some(BinOp {
+            apply: |a, b| a * b,
+            prio: 1,
+        }),
+        unary_op: None,
+    },
+    Operator {
+        repr: "/",
+        bin_op: Some(BinOp {
+            apply: |a, b| a / b,
+            prio: 1,
+        }),
+        unary_op: None,
+    },
+    Operator {
+        repr: "+",
+        bin_op: Some(BinOp {
+            apply: |a, b| a + b,
+            prio: 0,
+        }),
+        unary_op: Some(|a| a),
+    },
+    Operator {
+        repr: "-",
+        bin_op: Some(BinOp {
+            apply: |a, b| a - b,
+            prio: 0,
+        }),
+        unary_op: Some(|a| (-a)),
+    },
+    Operator {
+        repr: "sin",
+        bin_op: None,
+        unary_op: Some(|a| a.sin()),
+    }
+);
 
-fn exmex_parse_optimized<'a>(strings: &'a [&str]) -> Vec<FlatEx<'a, f64>> {
-    let ops = vec![
-        Operator {
-            repr: "^",
-            bin_op: Some(BinOp {
-                apply: |a: f64, b| a.powf(b),
-                prio: 2,
-            }),
-            unary_op: None,
-        },
-        Operator {
-            repr: "*",
-            bin_op: Some(BinOp {
-                apply: |a, b| a * b,
-                prio: 1,
-            }),
-            unary_op: None,
-        },
-        Operator {
-            repr: "/",
-            bin_op: Some(BinOp {
-                apply: |a, b| a / b,
-                prio: 1,
-            }),
-            unary_op: None,
-        },
-        Operator {
-            repr: "+",
-            bin_op: Some(BinOp {
-                apply: |a, b| a + b,
-                prio: 0,
-            }),
-            unary_op: Some(|a| a),
-        },
-        Operator {
-            repr: "-",
-            bin_op: Some(BinOp {
-                apply: |a, b| a - b,
-                prio: 0,
-            }),
-            unary_op: Some(|a| (-a)),
-        },
-        Operator {
-            repr: "sin",
-            bin_op: None,
-            unary_op: Some(|a| a.sin()),
-        },
-    ];
+fn exmex_parse_optimized<'a>(strings: &'a [&str]) -> Vec<FlatEx<'a, f64, OnlyNeededOperators>> {
     strings
         .iter()
-        .map(|expr_str| exmex::parse(expr_str, &ops).unwrap())
+        .map(|expr_str| FlatEx::<f64, OnlyNeededOperators>::from_str(expr_str).unwrap())
         .collect::<Vec<_>>()
 }
 
@@ -159,6 +185,21 @@ fn exmex_bench_parse_optimized(c: &mut Criterion) {
     run_benchmark_parse(exmex_parse_optimized, "exmex_parse_optimized", c);
 }
 
+fn exmex_bench_deepex_parseval(c: &mut Criterion) {
+    fn func(s: &str) -> f64 {
+        exmex::eval_str(s).unwrap()
+    }
+    run_benchmark_parseval(func, "exmex deepex", c);
+}
+
+fn exmex_bench_flatex_parseval(c: &mut Criterion) {
+    fn func(s: &str) -> f64 {
+        let flatex = FlatEx::<f64>::from_str(s).unwrap();
+        flatex.eval(&[]).unwrap()
+    }
+    run_benchmark_parseval(func, "exmex flatex", c);
+}
+
 fn exmex_bench_eval(c: &mut Criterion) {
     let parsed_exprs = exmex_parse(&BENCH_EXPRESSIONS_STRS);
     let funcs = parsed_exprs
@@ -313,37 +354,52 @@ fn rsc_bench_eval(c: &mut Criterion) {
     run_benchmark(funcs, "rsc", c);
 }
 
-fn exmex_bench_serde(c: &mut Criterion) -> () {
-    for (expr_str, expr_name, ref_func) in izip!(
+#[cfg(feature = "serde")]
+fn run_benchmark_serialize<Ex: Serialize>(expr: &Ex, expr_name: &str, c: &mut Criterion) {
+    c.bench_function(format!("exmex_serde_ser {}", expr_name).as_str(), |b| {
+        b.iter(|| {
+            serde_json::to_string(black_box(&expr)).unwrap();
+        })
+    });
+}
+
+#[cfg(feature = "serde")]
+fn run_benchmark_deserialize<'de, Ex: Deserialize<'de>>(
+    expr_str: &'de str,
+    expr_name: &str,
+    c: &mut Criterion,
+) {
+    c.bench_function(format!("exmex_serde_de {}", expr_name).as_str(), |b| {
+        b.iter(|| {
+            serde_json::from_str::<Ex>(black_box(expr_str)).unwrap();
+        })
+    });
+}
+
+fn exmex_bench_serde(_c: &mut Criterion) -> () {
+    for (_expr_str, _expr_name) in izip!(
         BENCH_EXPRESSIONS_STRS.iter(),
-        BENCH_EXPRESSIONS_NAMES.iter(),
-        BENCH_EXPRESSIONS_REFS.iter()
+        BENCH_EXPRESSIONS_NAMES.iter()
     ) {
-        let flatex = exmex::parse_with_default_ops::<f64>(expr_str).unwrap();
-        let flatex_owned = OwnedFlatEx::from_flatex(flatex.clone());
-        c.bench_function(format!("exmex_serde {}", expr_name).as_str(), |b| {
-            b.iter(|| {
-                let serialized = serde_json::to_string(&flatex).unwrap();
-                let deserialized = serde_json::from_str::<FlatEx<f64>>(&serialized).unwrap();
-                let serialized_owned = serde_json::to_string(&flatex_owned).unwrap();
-                let deserialized_owned =
-                    serde_json::from_str::<OwnedFlatEx<f64>>(&serialized_owned).unwrap();
-                let vars = &[BENCH_Y + BENCH_Z, BENCH_Y, BENCH_Z];
-                assert_float_eq(
-                    deserialized.eval(vars).unwrap(),
-                    ref_func(vars[0], vars[1], vars[2]),
-                );
-                assert_float_eq(
-                    deserialized_owned.eval(vars).unwrap(),
-                    ref_func(vars[0], vars[1], vars[2]),
-                );
-            })
-        });
+        #[cfg(feature = "serde")]
+        {
+            let expr_str_de = format!("\"{}\"", _expr_str);
+            let flatex = FlatEx::<f64>::from_str(_expr_str).unwrap();
+            let flatex_owned = OwnedFlatEx::from_flatex(flatex.clone());
+            let expr_name_ = format!("flatex {}", _expr_name);
+            run_benchmark_serialize(&flatex, &expr_name_, _c);
+            run_benchmark_deserialize::<FlatEx<f64>>(&expr_str_de, &expr_name_, _c);
+            let expr_name_ = format!("owned_flatex {}", _expr_name);
+            run_benchmark_serialize(&flatex_owned, &expr_name_, _c);
+            run_benchmark_deserialize::<OwnedFlatEx<f64>>(&expr_str_de, &expr_name_, _c);
+        }
     }
 }
 
 criterion_group!(
     benches,
+    exmex_bench_deepex_parseval,
+    exmex_bench_flatex_parseval,
     exmex_bench_serde,
     fasteval_bench_eval,
     exmex_bench_eval,
