@@ -1,11 +1,13 @@
 use crate::{
-    definitions::{N_NODES_ON_STACK, N_VARS_ON_STACK},
-    expression::deep_details::{
-        self, BinOpsWithReprsBuf, OverloadedOps, UnaryOpWithReprsBuf, ADD_REPR, DIV_REPR, MUL_REPR,
-        SUB_REPR,
+    definitions::{
+        N_BINOPS_OF_DEEPEX_ON_STACK, N_NODES_ON_STACK, N_UNARYOPS_OF_DEEPEX_ON_STACK,
+        N_VARS_ON_STACK,
     },
-    operators::{self, BinOp, UnaryOp},
-    parser, ExParseError, Operator,
+    expression::{
+        deep_details::{self, prioritized_indices, BinOpsWithReprsBuf, UnaryOpWithReprsBuf}
+    },
+    operators::{BinOp, DefaultOpsFactory, MakeOperators, UnaryOp},
+    parser, ExError, ExResult, Operator,
 };
 use num::Float;
 use regex::Regex;
@@ -14,7 +16,6 @@ use std::{
     fmt,
     fmt::{Debug, Display, Formatter},
     iter,
-    ops::{Add, Div, Mul, Sub},
     str::FromStr,
 };
 
@@ -23,6 +24,20 @@ pub type ExprIdxVec = SmallVec<[usize; N_NODES_ON_STACK]>;
 /// Container of binary operators of one expression.
 pub type BinOpVec<T> = SmallVec<[BinOp<T>; N_NODES_ON_STACK]>;
 
+pub fn parse<'a, T, F>(text: &'a str, ops: &[Operator<'a, T>], is_numeric: F) -> ExResult<DeepEx<'a, T>>
+where
+    T: Copy + Debug + FromStr,
+    <T as FromStr>::Err: Debug,
+    F: Fn(&'a str) -> Option<&'a str>,
+{
+    let parsed_tokens = parser::tokenize_and_analyze(text, ops, is_numeric)?;
+    parser::check_parsed_token_preconditions(&parsed_tokens)?;
+    let parsed_vars = deep_details::find_parsed_vars(&parsed_tokens);
+    let (expr, _) =
+        deep_details::make_expression(&parsed_tokens[0..], &parsed_vars, UnaryOpWithReprs::new())?;
+    Ok(expr)
+}
+
 /// A deep node can be an expression, a number, or
 /// a variable.
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd)]
@@ -43,6 +58,9 @@ where
     pub fn one() -> Self {
         DeepNode::Num(T::from(1.0).unwrap())
     }
+    pub fn num(n: T) -> Self {
+        DeepNode::Num(n)
+    }
 }
 impl<'a, T: Copy + Debug> Debug for DeepNode<'a, T> {
     fn fmt(&self, f: &mut Formatter) -> fmt::Result {
@@ -55,13 +73,13 @@ impl<'a, T: Copy + Debug> Debug for DeepNode<'a, T> {
 }
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub struct BinOpsWithReprs<'a, T: Copy> {
-    pub reprs: Vec<&'a str>,
+    pub reprs: SmallVec<[&'a str; N_BINOPS_OF_DEEPEX_ON_STACK]>,
     pub ops: BinOpVec<T>,
 }
 impl<'a, T: Copy> BinOpsWithReprs<'a, T> {
-    pub fn new() -> BinOpsWithReprs<'a, T> {
+    pub fn new() -> Self {
         BinOpsWithReprs {
-            reprs: vec![],
+            reprs: smallvec![],
             ops: BinOpVec::new(),
         }
     }
@@ -69,13 +87,13 @@ impl<'a, T: Copy> BinOpsWithReprs<'a, T> {
 
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
 pub struct UnaryOpWithReprs<'a, T: Copy> {
-    pub reprs: Vec<&'a str>,
+    pub reprs: SmallVec<[&'a str; N_UNARYOPS_OF_DEEPEX_ON_STACK]>,
     pub op: UnaryOp<T>,
 }
 impl<'a, T: Copy> UnaryOpWithReprs<'a, T> {
     pub fn new() -> UnaryOpWithReprs<'a, T> {
         UnaryOpWithReprs {
-            reprs: vec![],
+            reprs: smallvec![],
             op: UnaryOp::new(),
         }
     }
@@ -102,15 +120,14 @@ pub struct DeepEx<'a, T: Copy + Debug> {
     /// Unary operators are applied to the result of evaluating all nodes with all
     /// binary operators.
     unary_op: UnaryOpWithReprs<'a, T>,
-    overloaded_ops: Option<OverloadedOps<'a, T>>,
     var_names: SmallVec<[&'a str; N_VARS_ON_STACK]>,
 }
 
 fn lift_nodes<'a, T: Copy + Debug>(deepex: &mut DeepEx<'a, T>) {
     if deepex.nodes.len() == 1 && deepex.unary_op.op.len() == 0 {
-        match deepex.nodes[0].clone() {
+        match &deepex.nodes[0] {
             DeepNode::Expr(e) => {
-                *deepex = e;
+                *deepex = e.clone();
             }
             _ => (),
         }
@@ -164,7 +181,7 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
             }
         }
 
-        let mut resulting_reprs = vec![];
+        let mut resulting_reprs = smallvec![];
         self.bin_ops.ops = self
             .bin_ops
             .ops
@@ -195,9 +212,9 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
         nodes: Vec<DeepNode<'a, T>>,
         bin_ops: BinOpsWithReprs<'a, T>,
         unary_op: UnaryOpWithReprs<'a, T>,
-    ) -> Result<DeepEx<'a, T>, ExParseError> {
+    ) -> ExResult<DeepEx<'a, T>> {
         if nodes.len() != bin_ops.ops.len() + 1 {
-            Err(ExParseError {
+            Err(ExError {
                 msg: format!(
                     "mismatch between number of nodes {:?} and binary operators {:?} ({} vs {})",
                     nodes,
@@ -230,7 +247,6 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
                 nodes,
                 bin_ops,
                 unary_op,
-                overloaded_ops: None,
                 var_names: found_vars,
             };
             expr.compile();
@@ -238,145 +254,45 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
         }
     }
 
-    pub fn unparse(&self) -> String {
-        let mut node_strings = self.nodes.iter().map(|n| match n {
-            DeepNode::Num(n) => format!("{:?}", n),
-            DeepNode::Var((_, var_name)) => format!("{{{}}}", var_name),
-            DeepNode::Expr(e) => {
-                if e.unary_op.op.len() == 0 {
-                    format!("({})", e.unparse())
-                } else {
-                    e.unparse()
-                }
-            }
-        });
-        let mut bin_op_strings = self.bin_ops.reprs.iter();
-        // a valid expression has at least one node
-        let first_node_str = node_strings.next().unwrap();
-        let node_with_bin_ops_string = node_strings.fold(first_node_str, |mut res, node_str| {
-            let bin_op_str = bin_op_strings.next().unwrap();
-            res.push_str(bin_op_str);
-            res.push_str(node_str.as_str());
-            res
-        });
-        let unary_op_string = self
-            .unary_op
-            .reprs
-            .iter()
-            .fold(String::new(), |mut res, uop_str| {
-                res.push_str(uop_str);
-                res.push('(');
-                res
-            });
-        let closings = iter::repeat(")").take(self.unary_op.op.len()).fold(
-            String::new(),
-            |mut res, closing| {
-                res.push_str(closing);
-                res
-            },
-        );
-        if self.unary_op.op.len() == 0 {
-            node_with_bin_ops_string
-        } else {
-            format!(
-                "{}{}{}",
-                unary_op_string, node_with_bin_ops_string, closings
-            )
-        }
+    pub fn from_node(node: DeepNode<'a, T>) -> DeepEx<'a, T> {
+        DeepEx::new(vec![node], BinOpsWithReprs::new(), UnaryOpWithReprs::new()).unwrap()
     }
 
-    pub fn from_node(node: DeepNode<'a, T>, overloaded_ops: OverloadedOps<'a, T>) -> DeepEx<'a, T> {
-        let mut deepex =
-            DeepEx::new(vec![node], BinOpsWithReprs::new(), UnaryOpWithReprs::new()).unwrap();
-        deepex.set_overloaded_ops(Some(overloaded_ops));
-        deepex
-    }
-
-    pub fn one(overloaded_ops: OverloadedOps<'a, T>) -> DeepEx<'a, T>
+    pub fn one() -> DeepEx<'a, T>
     where
         T: Float,
     {
-        DeepEx::from_node(DeepNode::one(), overloaded_ops)
+        DeepEx::from_node(DeepNode::one())
     }
 
-    pub fn one_like(other: &DeepEx<'a, T>) -> Result<DeepEx<'a, T>, ExParseError>
+    pub fn zero() -> DeepEx<'a, T>
     where
         T: Float,
     {
-        Ok(DeepEx::one(other.unpack_and_clone_overloaded_ops()?))
+        DeepEx::from_node(DeepNode::zero())
     }
 
-    pub fn zero(overloaded_ops: OverloadedOps<'a, T>) -> DeepEx<'a, T>
+    pub fn from_num(x: T) -> DeepEx<'a, T>
     where
         T: Float,
     {
-        DeepEx::from_node(DeepNode::zero(), overloaded_ops)
+        DeepEx::from_node(DeepNode::num(x))
     }
 
     pub fn with_new_unary_op(self, unary_op: UnaryOpWithReprs<'a, T>) -> Self {
         Self {
             nodes: self.nodes,
-            overloaded_ops: self.overloaded_ops,
             bin_ops: self.bin_ops,
             var_names: self.var_names,
             unary_op,
         }
     }
 
-    pub fn from_str(text: &'a str) -> Result<DeepEx<'a, T>, ExParseError>
-    where
-        <T as std::str::FromStr>::Err: Debug,
-        T: Float + FromStr,
-    {
-        let ops = operators::make_default_operators::<T>();
-        DeepEx::from_ops(text, &ops)
-    }
-
-    pub fn from_ops(text: &'a str, ops: &[Operator<'a, T>]) -> Result<DeepEx<'a, T>, ExParseError>
-    where
-        <T as std::str::FromStr>::Err: Debug,
-        T: Copy + FromStr + Debug,
-    {
-        let parsed_tokens = parser::tokenize_and_analyze(text, ops, parser::is_numeric_text)?;
-        let mut deepex = deep_details::parsed_tokens_to_deepex(&parsed_tokens)?;
-        deepex.set_overloaded_ops(deep_details::find_overloaded_ops(ops));
-        Ok(deepex)
-    }
-
-    pub fn from_pattern(
-        text: &'a str,
-        ops: &[Operator<'a, T>],
-        number_regex_pattern: &str,
-    ) -> Result<DeepEx<'a, T>, ExParseError>
-    where
-        <T as std::str::FromStr>::Err: Debug,
-        T: Copy + FromStr + Debug,
-    {
-        let beginning_number_regex_regex = format!("^({})", number_regex_pattern);
-        let re_number = match Regex::new(beginning_number_regex_regex.as_str()) {
-            Ok(regex) => regex,
-            Err(_) => {
-                return Err(ExParseError {
-                    msg: "Cannot compile the passed number regex.".to_string(),
-                })
-            }
-        };
-        let is_numeric = |text: &'a str| parser::is_numeric_regex(&re_number, text);
-        let parsed_tokens = parser::tokenize_and_analyze(text, ops, is_numeric)?;
-        let mut deepex = deep_details::parsed_tokens_to_deepex(&parsed_tokens)?;
-        deepex.set_overloaded_ops(deep_details::find_overloaded_ops(ops));
-        Ok(deepex)
-    }
-
-    pub fn set_overloaded_ops(&mut self, ops: Option<OverloadedOps<'a, T>>) {
-        self.overloaded_ops = ops;
-    }
-
-    pub fn bin_ops(&self) -> &BinOpsWithReprs<T> {
+    pub fn bin_ops(&self) -> &BinOpsWithReprs<'a, T> {
         &self.bin_ops
     }
 
-    pub fn unary_op(&self) -> &UnaryOpWithReprs<T> {
+    pub fn unary_op(&self) -> &UnaryOpWithReprs<'a, T> {
         &self.unary_op
     }
 
@@ -384,12 +300,6 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
         &self.nodes
     }
 
-    pub fn unpack_and_clone_overloaded_ops(&self) -> Result<OverloadedOps<'a, T>, ExParseError> {
-        self.overloaded_ops.clone().ok_or(ExParseError {
-            msg: "cannot unpack overloaded ops when there are none".to_string(),
-        })
-    }
-
     fn is_num(&self, num: T) -> bool
     where
         T: Float,
@@ -457,8 +367,6 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
 
     /// Applies a binary operator to self and other
     pub fn operate_bin(self, other: Self, bin_op: BinOpsWithReprs<'a, T>) -> Self {
-        let overloaded_ops = self.overloaded_ops.clone();
-
         let (self_vars_updated, other_vars_updated) = self.var_names_union(other);
         let mut resex = DeepEx::new(
             vec![
@@ -469,7 +377,6 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
             UnaryOpWithReprs::new(),
         )
         .unwrap();
-        resex.overloaded_ops = overloaded_ops;
         resex.compile();
         resex
     }
@@ -481,60 +388,130 @@ impl<'a, T: Copy + Debug> DeepEx<'a, T> {
         self
     }
 
-    /// Applies one of the binary overloaded operators to self and other.
-    ///
-    /// # Panics
-    ///
-    /// if an overloaded operator has not been defined
-    ///
-    pub fn operate_overloaded(self, other: Self, repr: &'a str) -> Self {
-        if self.overloaded_ops.is_none() {
-            panic!("overloaded operators not available");
+    pub fn unparse_raw(&self) -> String {
+        let mut node_strings = self.nodes.iter().map(|n| match n {
+            DeepNode::Num(n) => format!("{:?}", n),
+            DeepNode::Var((_, var_name)) => format!("{{{}}}", var_name),
+            DeepNode::Expr(e) => {
+                if e.unary_op.op.len() == 0 {
+                    format!("({})", e.unparse_raw())
+                } else {
+                    e.unparse_raw()
+                }
+            }
+        });
+        let mut bin_op_strings = self.bin_ops.reprs.iter();
+        // a valid expression has at least one node
+        let first_node_str = node_strings.next().unwrap();
+        let node_with_bin_ops_string = node_strings.fold(first_node_str, |mut res, node_str| {
+            let bin_op_str = bin_op_strings.next().unwrap();
+            res.push_str(bin_op_str);
+            res.push_str(node_str.as_str());
+            res
+        });
+        let unary_op_string = self
+            .unary_op
+            .reprs
+            .iter()
+            .fold(String::new(), |mut res, uop_str| {
+                res.push_str(uop_str);
+                res.push('(');
+                res
+            });
+        let closings = iter::repeat(")").take(self.unary_op.op.len()).fold(
+            String::new(),
+            |mut res, closing| {
+                res.push_str(closing);
+                res
+            },
+        );
+        if self.unary_op.op.len() == 0 {
+            node_with_bin_ops_string
+        } else {
+            format!(
+                "{}{}{}",
+                unary_op_string, node_with_bin_ops_string, closings
+            )
         }
-        let overloaded_ops = self.overloaded_ops.clone();
-        let op = overloaded_ops.clone().unwrap().by_repr(repr);
+    }
 
-        let ops = smallvec![op.bin_op.unwrap()];
+    pub fn from_str_float(text: &'a str) -> ExResult<DeepEx<'a, T>>
+    where
+        <T as std::str::FromStr>::Err: Debug,
+        T: Float + FromStr,
+    {
+        let ops = DefaultOpsFactory::<T>::make();
+        DeepEx::from_ops(text, &ops)
+    }
 
-        let bin_op = BinOpsWithReprs {
-            reprs: vec![repr],
-            ops,
+
+    pub fn from_ops(text: &'a str, ops: &[Operator<'a, T>]) -> ExResult<DeepEx<'a, T>>
+    where
+        <T as std::str::FromStr>::Err: Debug,
+        T: Copy + FromStr + Debug,
+    {
+        parse(text, ops, parser::is_numeric_text)
+    }
+
+    pub fn from_pattern(
+        text: &'a str,
+        ops: &[Operator<'a, T>],
+        number_regex_pattern: &str,
+    ) -> ExResult<DeepEx<'a, T>>
+    where
+        <T as std::str::FromStr>::Err: Debug,
+        T: Copy + FromStr + Debug,
+    {
+        let beginning_num_re_pattern = format!("^({})", number_regex_pattern);
+        let re_number = match Regex::new(beginning_num_re_pattern.as_str()) {
+            Ok(regex) => regex,
+            Err(_) => {
+                return Err(ExError {
+                    msg: "Cannot compile the passed number regex.".to_string(),
+                })
+            }
         };
-        self.operate_bin(other, bin_op)
+        let is_numeric = |text: &'a str| parser::is_numeric_regex(&re_number, text);
+        parse(text, ops, is_numeric)
     }
-}
 
-impl<'a, T: Copy + Debug> Add for DeepEx<'a, T> {
-    type Output = Self;
-    fn add(self, other: Self) -> Self {
-        self.operate_overloaded(other, ADD_REPR)
+    pub fn eval(&self, vars: &[T]) -> ExResult<T> {
+        let mut numbers = self
+            .nodes
+            .iter()
+            .map(|node| -> ExResult<T> {
+                match node {
+                    DeepNode::Num(n) => Ok(*n),
+                    DeepNode::Var((idx, _)) => Ok(vars[*idx]),
+                    DeepNode::Expr(e) => e.eval(vars),
+                }
+            })
+            .collect::<ExResult<SmallVec<[T; N_NODES_ON_STACK]>>>()?;
+        let mut ignore: SmallVec<[bool; N_NODES_ON_STACK]> = smallvec![false; self.nodes.len()];
+        let prio_indices = prioritized_indices(&self.bin_ops.ops, &self.nodes);
+        for (i, &bin_op_idx) in prio_indices.iter().enumerate() {
+            let num_idx = prio_indices[i];
+            let mut shift_left = 0usize;
+            while ignore[num_idx - shift_left] {
+                shift_left += 1usize;
+            }
+            let mut shift_right = 1usize;
+            while ignore[num_idx + shift_right] {
+                shift_right += 1usize;
+            }
+            let num_1 = numbers[num_idx - shift_left];
+            let num_2 = numbers[num_idx + shift_right];
+            numbers[num_idx - shift_left] = (self.bin_ops.ops[bin_op_idx].apply)(num_1, num_2);
+            ignore[num_idx + shift_right] = true;
+        }
+        Ok(self.unary_op.op.apply(numbers[0]))
     }
-}
 
-impl<'a, T: Copy + Debug> Sub for DeepEx<'a, T> {
-    type Output = Self;
-    fn sub(self, other: Self) -> Self {
-        self.operate_overloaded(other, SUB_REPR)
-    }
-}
-
-impl<'a, T: Copy + Debug> Mul for DeepEx<'a, T> {
-    type Output = Self;
-    fn mul(self, other: Self) -> Self {
-        self.operate_overloaded(other, MUL_REPR)
-    }
-}
-
-impl<'a, T: Copy + Debug> Div for DeepEx<'a, T> {
-    type Output = Self;
-    fn div(self, other: Self) -> Self {
-        self.operate_overloaded(other, DIV_REPR)
-    }
 }
 
 impl<'a, T: Copy + Debug> Display for DeepEx<'a, T> {
     fn fmt(&self, f: &mut Formatter) -> fmt::Result {
-        write!(f, "{}", self.unparse())
+        write!(f, "{}", self.unparse_raw())
     }
 }
 #[derive(Clone, Eq, PartialEq, Ord, PartialOrd, Debug)]
@@ -571,26 +548,25 @@ impl<'a, T: Copy + Debug> DeepBuf<T> {
                 .collect(),
             bin_ops: BinOpsWithReprsBuf::from_deepex(deepex.bin_ops()),
             unary_op: UnaryOpWithReprsBuf::from_deepex(deepex.unary_op()),
-            unparsed: deepex.unparse(),
+            unparsed: deepex.unparse_raw(),
             var_names: deepex.var_names.iter().map(|vn| vn.to_string()).collect(),
         }
     }
-    pub fn to_deepex(&'a self, ops: &[Operator<'a, T>]) -> Result<DeepEx<'a, T>, ExParseError> {
+    pub fn to_deepex(&'a self, ops: &[Operator<'a, T>]) -> ExResult<DeepEx<'a, T>> {
         let mut deepex = DeepEx::new(
             self.nodes
                 .iter()
-                .map(|node| -> Result<_, ExParseError> {
+                .map(|node| -> ExResult<_> {
                     match node {
                         DeepBufNode::Expr(e) => Ok(DeepNode::Expr(e.to_deepex(ops)?)),
                         DeepBufNode::Num(n) => Ok(DeepNode::Num(*n)),
                         DeepBufNode::Var(v) => Ok(DeepNode::Var((v.0, v.1.as_str()))),
                     }
                 })
-                .collect::<Result<_, ExParseError>>()?,
+                .collect::<ExResult<_>>()?,
             self.bin_ops.to_deepex(),
             self.unary_op.to_deepex(),
         )?;
-        deepex.set_overloaded_ops(deep_details::find_overloaded_ops(ops));
         deepex.var_names = self.var_names.iter().map(|vn| vn.as_str()).collect();
         Ok(deepex)
     }
@@ -598,10 +574,8 @@ impl<'a, T: Copy + Debug> DeepBuf<T> {
 
 #[cfg(test)]
 use {
-    super::flat::flatten,
     crate::{
-        expression::partial_derivatives::partial_deepex,
-        operators::make_default_operators,
+        expression::{partial_derivatives::partial_deepex},
         util::{assert_float_eq, assert_float_eq_f64},
     },
     rand::{thread_rng, Rng},
@@ -610,12 +584,12 @@ use {
 
 #[test]
 fn test_reset_vars() {
-    let deepex = DeepEx::<f64>::from_str("2*z+x+y * .5").unwrap();
+    let deepex = DeepEx::<f64>::from_str_float("2*z+x+y * .5").unwrap();
     let ref_vars = ["x", "y", "z"];
     for i in 0..ref_vars.len() {
         assert_eq!(deepex.var_names[i], ref_vars[i]);
     }
-    let deepex2 = DeepEx::<f64>::from_str("a*c*b").unwrap();
+    let deepex2 = DeepEx::<f64>::from_str_float("a*c*b").unwrap();
     let ref_vars = ["a", "b", "c"];
     for i in 0..ref_vars.len() {
         assert_eq!(deepex2.var_names[i], ref_vars[i]);
@@ -626,17 +600,13 @@ fn test_reset_vars() {
         assert_eq!(deepex_.var_names[i], all_vars[i]);
         assert_eq!(deepex2_.var_names[i], all_vars[i]);
     }
-    assert_eq!(deepex.unparse(), deepex_.unparse());
-    assert_eq!(deepex2.unparse(), deepex2_.unparse());
-    let flatex = flatten(deepex);
-    let flatex2 = flatten(deepex2);
-    let flatex_ = flatten(deepex_);
-    let flatex2_ = flatten(deepex2_);
-    assert_float_eq_f64(flatex.eval(&[2.0, 6.0, 1.5]).unwrap(), 8.0);
-    assert_float_eq_f64(flatex2.eval(&[3.0, 5.0, 4.0]).unwrap(), 60.0);
-    assert_float_eq_f64(flatex_.eval(&[3.0, 5.0, 4.0, 2.0, 6.0, 1.5]).unwrap(), 8.0);
+    assert_eq!(deepex.unparse_raw(), deepex_.unparse_raw());
+    assert_eq!(deepex2.unparse_raw(), deepex2_.unparse_raw());
+    assert_float_eq_f64(deepex.eval(&[2.0, 6.0, 1.5]).unwrap(), 8.0);
+    assert_float_eq_f64(deepex2.eval(&[3.0, 5.0, 4.0]).unwrap(), 60.0);
+    assert_float_eq_f64(deepex_.eval(&[3.0, 5.0, 4.0, 2.0, 6.0, 1.5]).unwrap(), 8.0);
     assert_float_eq_f64(
-        flatex2_.eval(&[3.0, 5.0, 4.0, 2.0, 6.0, 1.5]).unwrap(),
+        deepex2_.eval(&[3.0, 5.0, 4.0, 2.0, 6.0, 1.5]).unwrap(),
         60.0,
     );
 }
@@ -644,7 +614,7 @@ fn test_reset_vars() {
 #[test]
 fn test_var_name_union() {
     fn from_str(text: &str) -> DeepEx<f64> {
-        DeepEx::from_str(text).unwrap()
+        DeepEx::from_str_float(text).unwrap()
     }
     fn test(str_1: &str, str_2: &str, var_names: &[&str]) {
         let first = from_str(str_1);
@@ -665,66 +635,11 @@ fn test_var_name_union() {
     test("x+y*z", "z+y", &vec!["x", "y", "z"]);
 }
 
-#[test]
-fn test_operator_overloading() {
-    fn from_str(text: &str) -> DeepEx<f64> {
-        DeepEx::from_str(text).unwrap()
-    }
-    fn eval<'a>(deepex: &DeepEx<'a, f64>, vars: &[f64], val: f64) {
-        assert_float_eq_f64(flatten(deepex.clone()).eval(vars).unwrap(), val);
-    }
-
-    fn check_shape<'a>(deepex: &DeepEx<'a, f64>, n_nodes: usize) {
-        assert_eq!(deepex.nodes.len(), n_nodes);
-        assert_eq!(deepex.bin_ops.ops.len(), n_nodes - 1);
-        assert_eq!(deepex.bin_ops.reprs.len(), n_nodes - 1);
-    }
-
-    let one = from_str("1");
-    let two = one.clone() + one.clone();
-    check_shape(&two, 1);
-    eval(&two, &[], 2.0);
-
-    let minus_one = from_str("-1");
-    let one = minus_one.clone() * minus_one.clone();
-    check_shape(&one, 1);
-    eval(&one, &[], 1.0);
-
-    let x_squared = from_str("x*x");
-    check_shape(&x_squared, 2);
-    let two_x_squared = two.clone() * x_squared.clone();
-    check_shape(&two_x_squared, 2);
-    eval(&two_x_squared, &[0.0], 0.0);
-    eval(&two_x_squared, &[1.0], 2.0);
-    eval(&two_x_squared, &[2.0], 8.0);
-    eval(&two_x_squared, &[3.0], 18.0);
-    let some_expr = from_str("x") + from_str("x") * from_str("2") / from_str("x^(.5)");
-    check_shape(&some_expr, 2);
-    eval(&some_expr, &[4.0], 8.0);
-
-    let x_plus_y_plus_z = from_str("x+y+z");
-    check_shape(&x_plus_y_plus_z, 3);
-    let y_minus_z = from_str("y-z");
-    check_shape(&y_minus_z, 2);
-    let prod_of_above = x_plus_y_plus_z.clone() * y_minus_z.clone();
-    check_shape(&prod_of_above, 2);
-    eval(&prod_of_above, &[1.0, 4.0, 8.0], -52.0);
-    let div_of_above = x_plus_y_plus_z.clone() / y_minus_z.clone();
-    eval(&div_of_above, &[1.0, 4.0, 8.0], -3.25);
-    let sub_of_above = x_plus_y_plus_z.clone() - y_minus_z.clone();
-    eval(&sub_of_above, &[1.0, 4.0, 8.0], 17.0);
-    let add_of_above = x_plus_y_plus_z + y_minus_z.clone();
-    eval(&add_of_above, &[1.0, 4.0, 8.0], 9.0);
-    let x_plus_cossin_y_plus_z = from_str("x+cos(sin(y+z))");
-    let prod_of_above = x_plus_cossin_y_plus_z * y_minus_z;
-    eval(&prod_of_above, &[1.0, 4.0, 8.0], -7.4378625090980925);
-}
-
 #[test]
 fn test_partial_finite() {
-    let ops = make_default_operators::<f64>();
+    let ops = DefaultOpsFactory::<f64>::make();
     fn test<'a>(sut: &str, ops: &'a [Operator<'a, f64>], range: Range<f64>) {
-        let dut = DeepEx::<f64>::from_str(sut).unwrap();
+        let dut = DeepEx::<f64>::from_str_float(sut).unwrap();
         let n_vars = dut.n_vars();
         let step = 1e-5;
         let mut rng = thread_rng();
@@ -734,31 +649,33 @@ fn test_partial_finite() {
             "test_partial_finite - checking derivatives at {:?} for {}",
             x0s, sut
         );
-        println!("test_partial_finite - dut vars {:?}", dut.var_names);
         for (var_idx, var_name) in dut.var_names.iter().enumerate() {
             let x1s: Vec<f64> = x0s
                 .iter()
                 .enumerate()
                 .map(|(i, x0)| if i == var_idx { x0 + step } else { *x0 })
                 .collect();
-            let flat_dut = flatten(dut.clone());
-            let f0 = flat_dut.eval(&x0s).unwrap();
-            let f1 = flat_dut.eval(&x1s).unwrap();
+
+
+            let f0 = dut.eval(&x0s).unwrap();
+            let f1 = dut.eval(&x1s).unwrap();
             let finite_diff = (f1 - f0) / step;
             let deri = partial_deepex(var_idx, dut.clone(), &ops).unwrap();
-            let flat_deri = flatten(deri.clone()).eval(&x0s).unwrap();
+            let deri = deri.eval(&x0s).unwrap();
             println!(
                 "test_partial_finite -\n {} (derivative)\n {} (finite diff)",
-                flat_deri, finite_diff
+                deri, finite_diff
             );
             let msg = format!("sut {}, d_{} is {}", sut, var_name, deri);
             println!("test_partial_finite - {}", msg);
-            assert_float_eq::<f64>(flat_deri, finite_diff, 1e-5, 1e-3, msg.as_str());
+            assert_float_eq::<f64>(deri, finite_diff, 1e-5, 1e-3, msg.as_str());
         }
     }
-
-    test("z*sin(x)+cos(y)^(1 + x^2)/(sin(z))", &ops, 0.01..1.0);
-    test("sin(sin(x+z))", &ops, -10.0..10.0);
+    test("sqrt(x)", &ops, 0.0..10000.0);
+    test("asin(x)", &ops, -1.0..1.0);
+    test("acos(x)", &ops, -1.0..1.0);
+    test("atan(x)", &ops, -1.0..1.0);
+    test("1/x", &ops, -10.0..10.0);
     test("x^x", &ops, 0.01..2.0);
     test("x^y", &ops, 4.036286084344371..4.036286084344372);
     test("z+sin(x)+cos(y)", &ops, -1.0..1.0);
@@ -767,50 +684,44 @@ fn test_partial_finite() {
     test("sin(x-z)", &ops, -10.0..10.0);
     test("y-sin(x-z)", &ops, -10.0..10.0);
     test("(sin(x)^2)/x/4", &ops, -10.0..10.0);
-    test("1/x", &ops, -10.0..10.0);
     test("sin(y+x)/((x*2)/y)*(2*x)", &ops, -1.0..1.0);
+    test("z*sin(x)+cos(y)^(1 + x^2)/(sin(z))", &ops, 0.01..1.0);
     test("log(x^2)", &ops, 0.1..10.0);
     test("tan(x)", &ops, -1.0..1.0);
     test("tan(exp(x))", &ops, -1000.0..0.0);
     test("exp(y-x)", &ops, -1.0..1.0);
     test("sqrt(exp(y-x))", &ops, -1000.0..0.0);
-    test("sqrt(x)", &ops, 0.0..10000.0);
-    test("asin(x)", &ops, -1.0..1.0);
-    test("acos(x)", &ops, -1.0..1.0);
-    test("atan(x)", &ops, -1.0..1.0);
+    test("sin(sin(x+z))", &ops, -10.0..10.0);
     test("asin(sqrt(x+y))", &ops, 0.0..0.5);
-    test("sinh(x)", &ops, -1.0..1.0);
-    test("cosh(x)", &ops, -1.0..1.0);
-    test("tanh(x)", &ops, -1.0..1.0);
 }
 
 #[test]
 fn test_var_names() {
-    let deepex = DeepEx::<f64>::from_str("x+y+{x}+z*(-y)").unwrap();
+    let deepex = DeepEx::<f64>::from_str_float("x+y+{x}+z*(-y)").unwrap();
     let reference: SmallVec<[&str; N_VARS_ON_STACK]> = smallvec!["x", "y", "z"];
     assert_eq!(deepex.var_names, reference);
 }
 
 #[test]
 fn test_deep_compile() {
-    let ops = make_default_operators();
+    let ops = DefaultOpsFactory::make();
     let nodes = vec![DeepNode::Num(4.5), DeepNode::Num(0.5), DeepNode::Num(1.4)];
     let bin_ops = BinOpsWithReprs {
-        reprs: vec![ops[1].repr, ops[3].repr],
+        reprs: smallvec![ops[1].repr, ops[3].repr],
         ops: smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()],
     };
     let unary_op = UnaryOpWithReprs {
-        reprs: vec![ops[6].repr],
+        reprs: smallvec![ops[6].repr],
         op: UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]),
     };
     let deep_ex = DeepEx::new(nodes, bin_ops, unary_op).unwrap();
 
     let bin_ops = BinOpsWithReprs {
-        reprs: vec![ops[1].repr, ops[3].repr],
+        reprs: smallvec![ops[1].repr, ops[3].repr],
         ops: smallvec![ops[1].bin_op.unwrap(), ops[3].bin_op.unwrap()],
     };
     let unary_op = UnaryOpWithReprs {
-        reprs: vec![ops[6].repr],
+        reprs: smallvec![ops[6].repr],
         op: UnaryOp::from_vec(smallvec![ops[6].unary_op.unwrap()]),
     };
     let nodes = vec![
@@ -831,21 +742,21 @@ fn test_deep_compile() {
 #[test]
 fn test_deep_compile_2() {
     let deepex =
-        DeepEx::<f64>::from_str("(({x}^2.0)*(({x}^1.0)*2.0))+((({x}^1.0)*2.0)*({x}^2.0))").unwrap();
+        DeepEx::<f64>::from_str_float("(({x}^2.0)*(({x}^1.0)*2.0))+((({x}^1.0)*2.0)*({x}^2.0))").unwrap();
     println!("{}", deepex);
     assert_eq!(
         format!("{}", deepex),
         "(({x}^2.0)*(({x}^1.0)*2.0))+((({x}^1.0)*2.0)*({x}^2.0))"
     );
 
-    let deepex = DeepEx::<f64>::from_str("(((a+x^2*x^2)))").unwrap();
+    let deepex = DeepEx::<f64>::from_str_float("(((a+x^2*x^2)))").unwrap();
     println!("{}", deepex);
     assert_eq!(format!("{}", deepex), "{a}+{x}^2.0*{x}^2.0");
 
-    let deepex = DeepEx::<f64>::from_str("1+(((a+x^2*x^2)))").unwrap();
+    let deepex = DeepEx::<f64>::from_str_float("1+(((a+x^2*x^2)))").unwrap();
     println!("{}", deepex);
     assert_eq!(format!("{}", deepex), "1.0+({a}+{x}^2.0*{x}^2.0)");
-    let mut ddeepex = partial_deepex(1, deepex, &make_default_operators()).unwrap();
+    let mut ddeepex = partial_deepex(1, deepex, &DefaultOpsFactory::make()).unwrap();
     ddeepex.compile();
     println!("{}", ddeepex);
     assert_eq!(
