@@ -1,33 +1,36 @@
-#![doc(html_root_url = "https://docs.rs/exmex/0.9.4")]
+#![doc(html_root_url = "https://docs.rs/exmex/0.10.0")]
 //! Exmex is a fast, simple, and **ex**tendable **m**athematical **ex**pression evaluator
 //! with the ability to compute [partial derivatives](FlatEx::partial) of expressions.  
+//!
+//! The following snippet shows how to evaluate a string.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
 //! use exmex;
-//! assert!((exmex::eval_str("1.5 * ((cos(0) + 23.0) / 2.0)")? - 18.0).abs() < 1e-12);
+//! assert!((exmex::eval_str::<f64>("1.5 * ((cos(0) + 23.0) / 2.0)")? - 18.0).abs() < 1e-12);
 //! #
 //! #     Ok(())
 //! # }
 //! ```
 //! For floats, we have a list of predifined operators containing
 //! `^`, `*`, `/`, `+`, `-`, `sin`, `cos`, `tan`, `exp`, `log`, and `log2`. The full list is
-//! defined in [`make_default_operators`](make_default_operators).
+//! defined in [`DefaultOpsFactory`](DefaultOpsFactory). Library users can also create their
+//! own operators as shown below in the section about extendability.
 //!
 //! ## Variables
-//! For variables we can use strings that are not in the list of operators as shown in the following expression.
+//! To define variables we can use strings that are not in the list of operators as shown in the following expression.
 //! Additionally, variables should consist only of letters, numbers, and underscores. More precisely, they need to fit the
 //! regular expression
 //! ```r"^[a-zA-Z_]+[a-zA-Z_0-9]*"```.
-//! Variables' values are passed as slices to [`eval`](FlatEx::eval).
+//! Variables' values are passed as slices to [`eval`](Express::eval).
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex;
+//! use exmex::prelude::*;
 //! let to_be_parsed = "log(z) + 2* (-z^2 + sin(4*y))";
-//! let expr = exmex::parse_with_default_ops::<f64>(to_be_parsed)?;
+//! let expr = exmex::parse::<f64>(to_be_parsed)?;
 //! assert!((expr.eval(&[3.7, 2.5])? - 14.992794866624788 as f64).abs() < 1e-12);
 //! #
 //! #     Ok(())
@@ -41,89 +44,128 @@
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex;
+//! use exmex::prelude::*;
 //! let x = 2.1f64;
 //! let y = 0.1f64;
 //! let to_be_parsed = "log({x+y})";  // {x+y} is the name of one(!) variable 😕.
-//! let expr = exmex::parse_with_default_ops::<f64>(to_be_parsed)?;
+//! let expr = exmex::parse::<f64>(to_be_parsed)?;
 //! assert!((expr.eval(&[x+y])? - 2.2f64.ln()).abs() < 1e-12);
 //! #
 //! #     Ok(())
 //! # }
 //! ```
+//! The value returned by [`parse`](parse) implements the [`Express`](Express) trait
+//! and is an instance of the struct [`FlatEx`](FlatEx).
+//!
 //! ## Extendability
-//! Library users can define their own set of operators as shown in the following.
+//!
+//! How to use custom operators as well as custom data types of the operands even with 
+//! non-numeric literals is described in the following sub-sections.
+//!
+//! ### Custom Operators
+//!
+//! Operators are instances of the struct
+//! [`Operator`](Operator). An operator's representation in the string-to-be-parsed is defined in the field
+//! [`repr`](Operator::repr). Further, operator instances define a binary and a unary operator, since an operator 
+//! representation can correspond to both such as `-` or `+`. Note that we expect a unary operator to be always 
+//! on the left of a number.
+//!
+//! To make serialization via [`serde`](https://serde.rs/) possible, operators need to be created by factories as 
+//! shown in the following.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{BinOp, Operator};
-//! let ops = [
+//! use exmex::prelude::*;
+//! use exmex::{BinOp, MakeOperators, Operator, ops_factory};
+//! ops_factory!(
+//!     IntegerOpsFactory,  // name of the factory type
+//!     i32,                // data type of the operands
 //!     Operator {
 //!         repr: "%",
-//!         bin_op: Some(BinOp{ apply: |a: i32, b: i32| a % b, prio: 1 }),
+//!         bin_op: Some(BinOp{ apply: |a, b| a % b, prio: 1 }),
 //!         unary_op: None,
-//!     },
+//!         },
 //!     Operator {
 //!         repr: "/",
-//!         bin_op: Some(BinOp{ apply: |a: i32, b: i32| a / b, prio: 1 }),
+//!         bin_op: Some(BinOp{ apply: |a, b| a / b, prio: 1 }),
 //!         unary_op: None,
-//!     },
-//! ];
+//!     }
+//! );
 //! let to_be_parsed = "19 % 5 / 2 / a";
-//! let expr = exmex::parse::<i32>(to_be_parsed, &ops)?;
+//! let expr = FlatEx::<_, IntegerOpsFactory>::from_str(to_be_parsed)?;
 //! assert_eq!(expr.eval(&[1])?, 2);
 //! #
 //! #     Ok(())
 //! # }
 //! ```
 //!
-//! ### Operators
-//!
-//! Operators are instances of the struct
-//! [`Operator`](Operator) that has its representation in the field
-//! [`repr`](Operator::repr), a binary and a unary operator of
-//! type [`Option<BinOp<T>>`](Operator::bin_op) and
-//! [`Option<fn(T) -> T>`](Operator::unary_op), respectively, as
-//! members. [`BinOp`](BinOp)
-//! contains in addition to the function pointer [`apply`](BinOp::apply) of type `fn(T, T) -> T` an
-//! integer [`prio`](BinOp::prio). Operators
-//! can be both, binary and unary. See, e.g.,  `-` defined in the list of default
-//! operators. Note that we expect a unary operator to be always on the left of a
-//! number.
-//!
-//! ### Data Types of Numbers
+//! To extend an existing list of operators, the macro [`ops_factory`](ops_factory) is not 
+//! sufficient. In this case one has to create a factory struct and implement the 
+//! [`MakeOperators`](MakeOperators) trait with a little boilerplate code.
+//! ```rust
+//! # use std::error::Error;
+//! # fn main() -> Result<(), Box<dyn Error>> {
+//! #
+//! use exmex::prelude::*;
+//! use exmex::{BinOp, DefaultOpsFactory, MakeOperators, Operator};
+//! #[derive(Clone)]
+//! struct ExtendedOpsFactory;
+//! impl MakeOperators<f32> for ExtendedOpsFactory {
+//!     fn make<'a>() -> Vec<Operator<'a, f32>> {
+//!         let mut ops = DefaultOpsFactory::<f32>::make();
+//!         ops.push(
+//!             Operator {
+//!                 repr: "invert",
+//!                 bin_op: None,
+//!                 unary_op: Some(|a: f32| 1.0 / a),
+//!             },
+//!         );
+//!         ops
+//!     }
+//! }
+//! let to_be_parsed = "1 / a + invert(a)";
+//! let expr = FlatEx::<_, ExtendedOpsFactory>::from_str(to_be_parsed)?;
+//! assert!((expr.eval(&[3.0])? - 2.0/3.0).abs() < 1e-12);
+//! #
+//! #     Ok(())
+//! # }
+//! ```
+//! ### Custom Data Types of Numbers
 //!
 //! You can use any type that implements [`Copy`](core::marker::Copy) and
 //! [`FromStr`](std::str::FromStr). In case the representation of your data type in the
 //! string does not match the number regex `r"\.?[0-9]+(\.[0-9]+)?"`, you have to pass a
 //! suitable regex and use the function
-//! [`parse_with_number_pattern`](parse_with_number_pattern) instead of
-//! [`parse`](parse). Here is an example for `bool`.
+//! [`from_pattern`](Express::from_pattern) instead of
+//! [`from_str`](Express::from_str). Here is an example for `bool`.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{self, BinOp, Operator};
-//! let ops = [
+//! use exmex::prelude::*;
+//! use exmex::{BinOp, MakeOperators, Operator, ops_factory};
+//! ops_factory!(
+//!     BooleanOpsFactory, 
+//!     bool,
 //!     Operator {
 //!         repr: "&&",
-//!         bin_op: Some(BinOp{ apply: |a: bool, b: bool| a && b, prio: 1 }),
+//!         bin_op: Some(BinOp{ apply: |a, b| a && b, prio: 1 }),
 //!         unary_op: None,
 //!     },
 //!     Operator {
 //!         repr: "||",
-//!         bin_op: Some(BinOp{ apply: |a: bool, b: bool| a || b, prio: 1 }),
+//!         bin_op: Some(BinOp{ apply: |a, b| a || b, prio: 1 }),
 //!         unary_op: None,
 //!     },
 //!     Operator {
 //!         repr: "!",
 //!         bin_op: None,
-//!         unary_op: Some(|a: bool| !a),
-//!     },
-//! ];
+//!         unary_op: Some(|a| !a),
+//!     }
+//! );
 //! let to_be_parsed = "!(true && false) || (!false || (true && false))";
-//! let expr = exmex::parse_with_number_pattern::<bool>(to_be_parsed, &ops, "true|false")?;
+//! let expr = FlatEx::<_, BooleanOpsFactory>::from_pattern(to_be_parsed, "true|false")?;
 //! assert_eq!(expr.eval(&[])?, true);
 //! #
 //! #     Ok(())
@@ -133,13 +175,13 @@
 //! ## Partial Derivatives
 //!
 //! For default operators, expressions can be transformed into their partial derivatives
-//! again represented by expressions.
+//! again represented by expressions. To this end, there exists the method [`partial`](Express::partial).
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex;
-//! let expr = exmex::parse_with_default_ops::<f64>("x^2 + y^2")?;
+//! use exmex::prelude::*;
+//! let expr = exmex::parse::<f64>("x^2 + y^2")?;
 //! let dexpr_dx = expr.clone().partial(0)?;
 //! let dexpr_dy = expr.partial(1)?;
 //! assert!((dexpr_dx.eval(&[3.0, 2.0])? - 6.0).abs() < 1e-12);
@@ -148,44 +190,44 @@
 //! #     Ok(())
 //! # }
 //! ```
-//! 
+//!
 //! ## Owned Expression
-//! You cannot return a usual expression from a function without a lifetime parameter, 
-//! since expressions that are instances of [`FlatEx`](FlatEx) keep `&str`s instead of 
+//! You cannot return all expression types from a function without a lifetime parameter.
+//! For instance, expressions that are instances of [`FlatEx`](FlatEx) keep `&str`s instead of
 //! `String`s of variable or operator names to make faster parsing possible.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{self, ExParseError, FlatEx};
-//! fn make<'a>() -> Result<FlatEx::<'a, f64>, ExParseError> {
-//! //       |                        |
-//! //      lifetime parameter necessary
+//! use exmex::prelude::*;
+//! use exmex::ExResult;
+//! fn create_expr<'a>() -> ExResult<FlatEx::<'a, f64>> {
+//! //              |                          |
+//! //              lifetime parameter necessary
 //!
 //!     let to_be_parsed = "log(z) + 2* (-z^2 + sin(4*y))";
-//!     exmex::parse_with_default_ops::<f64>(to_be_parsed)
+//!     exmex::parse::<f64>(to_be_parsed)
 //! }
-//! let expr = make()?;
+//! let expr = create_expr()?;
 //! assert!((expr.eval(&[3.7, 2.5])? - 14.992794866624788 as f64).abs() < 1e-12);
 //! #
 //! #     Ok(())
 //! # }
-//! ``` 
-//! If you are willing to pay the price of roughly doubled parsing times, you can 
-//! obtain an expression that is an instance of [`OwnedFlatEx`](OwnedFlatEx) and owns 
+//! ```
+//! If you are willing to pay the price of roughly doubled parsing times, you can
+//! obtain an expression that is an instance of [`OwnedFlatEx`](OwnedFlatEx) and owns
 //! its strings. Evaluation times should be comparable. However, a lifetime parameter is
-//! not needed anymore as shown in the following. 
+//! not needed anymore as shown in the following.
 //! ```rust
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex::{self, ExParseError, OwnedFlatEx};
-//! fn make() -> Result<OwnedFlatEx::<f64>, ExParseError> {
+//! use exmex::{ExResult, Express, OwnedFlatEx};
+//! fn create_expr() -> ExResult<OwnedFlatEx::<f64>> {
 //!     let to_be_parsed = "log(z) + 2* (-z^2 + sin(4*y))";
-//!     let expr = exmex::parse_with_default_ops::<f64>(to_be_parsed)?;
-//!     Ok(OwnedFlatEx::from_flatex(expr))
+//!     OwnedFlatEx::<f64>::from_str(to_be_parsed)
 //! }
-//! let expr_owned = make()?;
+//! let expr_owned = create_expr()?;
 //! assert!((expr_owned.eval(&[3.7, 2.5])? - 14.992794866624788 as f64).abs() < 1e-12);
 //! #
 //! #     Ok(())
@@ -201,7 +243,7 @@
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
 //! use exmex;
-//! assert_eq!(exmex::eval_str("---1")?, -1.0);
+//! assert_eq!(exmex::eval_str::<f64>("---1")?, -1.0);
 //! #
 //! #     Ok(())
 //! # }
@@ -218,9 +260,9 @@
 //! # use std::error::Error;
 //! # fn main() -> Result<(), Box<dyn Error>> {
 //! #
-//! use exmex;
-//! let flatex = exmex::parse_with_default_ops::<f64>("-sin(z)/cos(mother_of_names) + 2^7")?;
-//! assert_eq!(format!("{}", flatex), "-(sin({z}))/cos({mother_of_names})+128.0");
+//! use exmex::prelude::*;
+//! let expr = exmex::parse::<f64>("-sin(z)/cos(mother_of_names) + 2^7")?;
+//! assert_eq!(format!("{}", expr), "-(sin({z}))/cos({mother_of_names})+128.0");
 //! #
 //! #     Ok(())
 //! # }
@@ -228,116 +270,71 @@
 //!
 //! ## Serialization and Deserialization
 //!
-//! To use [`serde`](https://serde.rs/) you can activate the feature `serde_support`.
-//! Currently, this only works for default operators. The implementation
-//! un-parses and re-parses the whole expression. 
-//! [`Deserialize`](https://docs.serde.rs/serde/de/trait.Deserialize.html) and 
+//! To use [`serde`](https://serde.rs/) you can activate the feature `serde`.
+//! The implementation un-parses and re-parses the whole expression.
+//! [`Deserialize`](https://docs.serde.rs/serde/de/trait.Deserialize.html) and
 //! [`Serialize`](https://docs.serde.rs/serde/de/trait.Serialize.html) are implemented for
-//! for both, [`FlatEx`](FlatEx) and [`OwnedFlatEx`](OwnedFlatEx).
+//! both, [`FlatEx`](FlatEx) and [`OwnedFlatEx`](OwnedFlatEx).
 //!
 //! ## Unicode
 //! Unicode input strings are currently not supported 😕 but might be added in the
 //! future 😀.
 //!
 
-mod definitions;
-mod expression;
-mod operators;
-mod parser;
-mod util;
-
 use std::{fmt::Debug, str::FromStr};
 
-pub use expression::flat::{FlatEx, OwnedFlatEx};
-use expression::{deep::DeepEx, flat};
-
+use expression::deep::DeepEx;
 use num::Float;
-pub use operators::{make_default_operators, BinOp, Operator};
-pub use parser::ExParseError;
+
+mod definitions;
+mod expression;
+#[macro_use]
+mod operators;
+mod parser;
+mod result;
+mod util;
+pub use {
+    expression::{
+        flat::{FlatEx, OwnedFlatEx},
+        Express,
+    },
+    operators::{BinOp, DefaultOpsFactory, MakeOperators, Operator},
+    result::{ExError, ExResult},
+};
+
+/// To use the expression trait [`Express`](Express) and its implementation [`FlatEx`](FlatEx)
+/// one can `use exmex::prelude::*;`.
+pub mod prelude {
+    pub use super::expression::{flat::FlatEx, Express};
+}
 
 /// Parses a string, evaluates a string, and returns the resulting number.
 ///
 /// # Errrors
 ///
 /// In case the parsing went wrong, e.g., due to an invalid input string, an
-/// [`ExParseError`](ExParseError) is returned.
+/// [`ExError`](ExError) is returned.
 ///
-pub fn eval_str(text: &str) -> Result<f64, ExParseError> {
-    let flatex = parse_with_default_ops(text)?;
-    flatex.eval(&[])
+pub fn eval_str<T: Float + FromStr + Debug>(text: &str) -> ExResult<T>
+where
+    <T as FromStr>::Err: Debug,
+{
+    let deepex = DeepEx::<T>::from_str_float(text)?;
+    deepex.eval(&[])
 }
 
-/// Parses a string and a vector of operators into an expression that can be evaluated.
+/// Parses a string and returns the expression that can be evaluated.
 ///
-/// # Errors
+/// # Errrors
 ///
-/// An error is returned in case [`parse_with_number_pattern`](parse_with_number_pattern)
-/// returns one.
-pub fn parse<'a, T>(text: &'a str, ops: &[Operator<'a, T>]) -> Result<FlatEx<'a, T>, ExParseError>
+/// In case the parsing went wrong, e.g., due to an invalid input string, an
+/// [`ExError`](ExError) is returned.
+///
+pub fn parse<T: Float + FromStr + Debug>(text: &str) -> ExResult<FlatEx<T>>
 where
-    <T as std::str::FromStr>::Err: Debug,
-    T: Copy + FromStr + Debug,
+    <T as FromStr>::Err: Debug,
 {
-    let deepex = DeepEx::from_ops(text, ops)?;
-    Ok(flat::flatten(deepex))
-}
-
-/// Parses a string and a vector of operators and a regex pattern that defines the looks
-/// of a number into an expression that can be evaluated.
-///
-/// # Errors
-///
-/// An [`ExParseError`](ExParseError) is returned, if
-///
-//
-// from apply_regexes
-//
-/// * the argument `number_regex_pattern` cannot be compiled,
-/// * the argument `text` contained a character that did not match any regex (e.g.,
-///   if there is a `Δ` in `text` but no [operator](Operator) with
-///   [`repr`](Operator::repr) equal to `Δ` is given),
-//
-// from check_preconditions
-//
-/// * the to-be-parsed string is empty,
-/// * a number or variable is next to another one, e.g., `2 {x}`,
-/// * wlog a number or variable is on the right of a closing parenthesis, e.g., `)5`,
-/// * a binary operator is next to another binary operator, e.g., `2*/4`,
-/// * wlog a closing parenthesis is next to an opening one, e.g., `)(` or `()`,
-/// * too many closing parentheses at some position, e.g., `(4+6) - 5)*2`,
-/// * the last element is an operator, e.g., `1+`,
-/// * the number of opening and closing parenthesis do not match, e.g., `((4-2)`,
-//
-// from make_expression
-//
-/// * in `parsed_tokens` a closing parentheses is directly following an operator, e.g., `+)`, or
-/// * a unary operator is followed directly by a binary operator, e.g., `sin*`.
-///
-pub fn parse_with_number_pattern<'a, T>(
-    text: &'a str,
-    ops: &[Operator<'a, T>],
-    number_regex_pattern: &str,
-) -> Result<FlatEx<'a, T>, ExParseError>
-where
-    <T as std::str::FromStr>::Err: Debug,
-    T: Copy + FromStr + Debug,
-{
-    let deepex = DeepEx::from_pattern(text, ops, number_regex_pattern)?;
-    Ok(flat::flatten(deepex))
-}
-
-/// Parses a string into an expression that can be evaluated using default operators.
-///
-/// # Errors
-///
-/// An error is returned in case [`parse`](parse)
-/// returns one.
-pub fn parse_with_default_ops<'a, T>(text: &'a str) -> Result<FlatEx<'a, T>, ExParseError>
-where
-    <T as std::str::FromStr>::Err: Debug,
-    T: Float + FromStr + Debug,
-{
-    Ok(flat::flatten(DeepEx::from_str(text)?))
+    FlatEx::<T>::from_str(text)
 }
 
 #[cfg(test)]
@@ -348,18 +345,19 @@ mod tests {
     use rand::Rng;
     use smallvec::{smallvec, SmallVec};
 
+    use crate::prelude::*;
     use crate::{
         eval_str,
-        operators::{make_default_operators, BinOp, Operator},
-        parse, parse_with_default_ops,
+        operators::{BinOp, DefaultOpsFactory, MakeOperators, Operator},
+        parse,
         util::{assert_float_eq_f32, assert_float_eq_f64},
-        ExParseError, FlatEx, OwnedFlatEx,
+        ExResult, OwnedFlatEx,
     };
 
     #[test]
     fn test_readme() {
-        fn readme_partial() -> Result<(), ExParseError> {
-            let expr = parse_with_default_ops::<f64>("y*x^2")?;
+        fn readme_partial() -> ExResult<()> {
+            let expr = parse::<f64>("y*x^2")?;
 
             // d_x
             let dexpr_dx = expr.partial(0)?;
@@ -377,20 +375,22 @@ mod tests {
 
             Ok(())
         }
-        fn readme() -> Result<(), ExParseError> {
+        fn readme() -> ExResult<()> {
             let result = eval_str("sin(73)")?;
             assert_float_eq_f64(result, 73f64.sin());
-            let expr = parse_with_default_ops::<f64>("2*x^3-4/z")?;
+            let expr = parse::<f64>("2*x^3-4/z")?;
             let value = expr.eval(&[5.3, 0.5])?;
             assert_float_eq_f64(value, 289.75399999999996);
             Ok(())
         }
-        fn readme_int() -> Result<(), ExParseError> {
-            let ops = vec![
+        fn readme_int() -> ExResult<()> {
+            ops_factory!(
+                BitwiseOpsFactory,
+                u32,
                 Operator {
                     repr: "|",
                     bin_op: Some(BinOp {
-                        apply: |a: u32, b: u32| a | b,
+                        apply: |a, b| a | b,
                         prio: 0,
                     }),
                     unary_op: None,
@@ -398,10 +398,10 @@ mod tests {
                 Operator {
                     repr: "!",
                     bin_op: None,
-                    unary_op: Some(|a: u32| !a),
-                },
-            ];
-            let expr = parse::<u32>("!(a|b)", &ops)?;
+                    unary_op: Some(|a| !a),
+                }
+            );
+            let expr = FlatEx::<_, BitwiseOpsFactory>::from_str("!(a|b)")?;
             let result = expr.eval(&[0, 1])?;
             assert_eq!(result, u32::MAX - 1);
             Ok(())
@@ -413,32 +413,32 @@ mod tests {
     #[test]
     fn test_variables_curly_space_names() {
         let sut = "{x } + { y }";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 1.0]).unwrap(), 2.0);
         assert_eq!(expr.unparse().unwrap(), "{x }+{ y }");
         let sut = "2*(4*{ xasd sa } + { y z}^2)";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.0]).unwrap(), 34.0);
         assert_eq!(expr.unparse().unwrap(), "2.0*(4.0*{ xasd sa }+{ y z}^2.0)");
     }
     #[test]
     fn test_variables_curly() {
         let sut = "5*{x} +  4*log2(log(1.5+{gamma}))*({x}*-(tan(cos(sin(652.2-{gamma}))))) + 3*{x}";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.2, 1.0]).unwrap(), 8.040556934857268);
 
         let sut = "sin({myvwmlf4i58eo;w/-sin(a)r_25})";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
         let sut = "((sin({myvar_25})))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
     }
     #[test]
     fn test_variables() {
         let sut = "sin  ({x})+(((cos({y})   ^  (sin({z})))*log(cos({y})))*cos({z}))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         let reference =
             |x: f64, y: f64, z: f64| x.sin() + y.cos().powf(z.sin()) * y.cos().ln() * z.cos();
 
@@ -449,8 +449,7 @@ mod tests {
         );
 
         let sut = "sin(sin(x - 1 / sin(y * 5)) + (5.0 - 1/z))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
-        let expr = OwnedFlatEx::from_flatex(expr);
+        let expr = OwnedFlatEx::<f64>::from_str(sut).unwrap();
         let reference =
             |x: f64, y: f64, z: f64| ((x - 1.0 / (y * 5.0).sin()).sin() + (5.0 - 1.0 / z)).sin();
         assert_float_eq_f64(
@@ -459,72 +458,68 @@ mod tests {
         );
 
         let sut = "0.02*sin( - (3*(2*(5.0 - 1/z))))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |z: f64| 0.02 * (-(3.0 * (2.0 * (5.0 - 1.0 / z)))).sin();
         assert_float_eq_f64(expr.eval(&[4.0]).unwrap(), reference(4.0));
 
         let sut = "y + 1 + 0.5 * x";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
-        let expr = OwnedFlatEx::from_flatex(expr);
+        let expr = OwnedFlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[3.0, 1.0]).unwrap(), 3.5);
 
         let sut = " -(-(1+x))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
-        let expr = OwnedFlatEx::from_flatex(expr);
+        let expr = OwnedFlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 2.0);
 
         let sut = " sin(cos(-3.14159265358979*x))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), -0.841470984807896);
 
         let sut = "5*sin(x * (4-y^(2-x) * 3 * cos(x-2*(y-1/(y-2*1/cos(sin(x*y))))))*x)";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
-        let expr = OwnedFlatEx::from_flatex(expr);
+        let expr = OwnedFlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5, 0.2532]).unwrap(), -3.1164569260604176);
 
         let sut = "5*x + 4*y + 3*x";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0, 0.0]).unwrap(), 8.0);
 
         let sut = "5*x + 4*y";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
-        let expr = OwnedFlatEx::from_flatex(expr);
+        let expr = OwnedFlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[0.0, 1.0]).unwrap(), 4.0);
 
         let sut = "5*x + 4*y + x^2";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.5, 3.7]).unwrap(), 33.55);
         assert_float_eq_f64(expr.eval(&[12.0, 9.3]).unwrap(), 241.2);
 
         let sut = "2*(4*x + y^2)";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.0]).unwrap(), 34.0);
 
         let sut = "sin(myvar_25)";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
         let sut = "((sin(myvar_25)))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.5707963267948966]).unwrap(), 1.0);
 
         let sut = "(0 * myvar_25 + cos(x))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(
             expr.eval(&[1.5707963267948966, 3.141592653589793]).unwrap(),
             -1.0,
         );
 
         let sut = "(-x^2)";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[1.0]).unwrap(), 1.0);
 
         let sut = "log(x) + 2* (-x^2 + sin(4*y))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.5, 3.7]).unwrap(), 14.992794866624788);
 
         let sut = "-sqrt(x)/(tanh(5-x)*2) + floor(2.4)* 1/asin(-x^2 + sin(4*sinh(y)))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(
             expr.eval(&[2.5, 3.7]).unwrap(),
             -(2.5f64.sqrt()) / (2.5f64.tanh() * 2.0)
@@ -532,87 +527,106 @@ mod tests {
         );
 
         let sut = "asin(sin(x)) + acos(cos(x)) + atan(tan(x))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[0.5]).unwrap(), 1.5);
 
         let sut = "sqrt(alpha^ceil(centauri))";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[2.0, 3.1]).unwrap(), 4.0);
 
         let sut = "trunc(x) + fract(x)";
-        let expr = parse_with_default_ops::<f64>(sut).unwrap();
+        let expr = FlatEx::<f64>::from_str(sut).unwrap();
         assert_float_eq_f64(expr.eval(&[23422.52345]).unwrap(), 23422.52345);
     }
 
     #[test]
     fn test_custom_ops_invert() {
-        let ops = vec![
-            Operator {
-                repr: "invert",
-                bin_op: None,
-                unary_op: Some(|a: f32| 1.0 / a),
-            },
-            Operator {
-                repr: "sqrt",
-                bin_op: None,
-                unary_op: Some(|a: f32| a.sqrt()),
-            },
-        ];
-        let expr = parse("sqrt(invert(a))", &ops).unwrap();
+        #[derive(Clone)]
+        struct SomeF32Operators;
+        impl MakeOperators<f32> for SomeF32Operators {
+            fn make<'a>() -> Vec<Operator<'a, f32>> {
+                vec![
+                    Operator {
+                        repr: "invert",
+                        bin_op: None,
+                        unary_op: Some(|a: f32| 1.0 / a),
+                    },
+                    Operator {
+                        repr: "sqrt",
+                        bin_op: None,
+                        unary_op: Some(|a: f32| a.sqrt()),
+                    },
+                ]
+            }
+        }
+        let expr = OwnedFlatEx::<f32, SomeF32Operators>::from_str("sqrt(invert(a))").unwrap();
         assert_float_eq_f32(expr.eval(&[0.25]).unwrap(), 2.0);
     }
 
     #[test]
     fn test_custom_ops() {
-        let custom_ops = vec![
-            Operator {
-                repr: "**",
-                bin_op: Some(BinOp {
-                    apply: |a: f32, b| a.powf(b),
-                    prio: 2,
-                }),
-                unary_op: None,
-            },
-            Operator {
-                repr: "*",
-                bin_op: Some(BinOp {
-                    apply: |a, b| a * b,
-                    prio: 1,
-                }),
-                unary_op: None,
-            },
-            Operator {
-                repr: "invert",
-                bin_op: None,
-                unary_op: Some(|a: f32| 1.0 / a),
-            },
-            Operator {
-                repr: "None",
-                bin_op: None,
-                unary_op: None,
-            },
-        ];
-        let expr = parse("2**2*invert(3)", &custom_ops).unwrap();
+        #[derive(Clone)]
+        struct SomeF32Operators;
+        impl MakeOperators<f32> for SomeF32Operators {
+            fn make<'a>() -> Vec<Operator<'a, f32>> {
+                vec![
+                    Operator {
+                        repr: "**",
+                        bin_op: Some(BinOp {
+                            apply: |a: f32, b| a.powf(b),
+                            prio: 2,
+                        }),
+                        unary_op: None,
+                    },
+                    Operator {
+                        repr: "*",
+                        bin_op: Some(BinOp {
+                            apply: |a, b| a * b,
+                            prio: 1,
+                        }),
+                        unary_op: None,
+                    },
+                    Operator {
+                        repr: "invert",
+                        bin_op: None,
+                        unary_op: Some(|a: f32| 1.0 / a),
+                    },
+                    Operator {
+                        repr: "None",
+                        bin_op: None,
+                        unary_op: None,
+                    },
+                ]
+            }
+        }
+        let expr = OwnedFlatEx::<f32, SomeF32Operators>::from_str("2**2*invert(3)").unwrap();
         let val = expr.eval(&[]).unwrap();
         assert_float_eq_f32(val, 4.0 / 3.0);
 
-        let expr = parse("None(5)", &custom_ops);
+        let expr = FlatEx::<f32, SomeF32Operators>::from_str("None(5)");
         assert!(expr.is_err());
 
-        let zero_mapper = Operator {
-            repr: "zer0",
-            bin_op: Some(BinOp {
-                apply: |_: f32, _| 0.0,
-                prio: 2,
-            }),
-            unary_op: Some(|_| 0.0),
-        };
-        let extended_operators = make_default_operators::<f32>()
-            .iter()
-            .cloned()
-            .chain(once(zero_mapper))
-            .collect::<Vec<_>>();
-        let expr = parse("2^2*1/(berti) + zer0(4)", &extended_operators).unwrap();
+        #[derive(Clone)]
+        struct ExtendedF32Operators;
+        impl MakeOperators<f32> for ExtendedF32Operators {
+            fn make<'a>() -> Vec<Operator<'a, f32>> {
+                let zero_mapper = Operator {
+                    repr: "zer0",
+                    bin_op: Some(BinOp {
+                        apply: |_: f32, _| 0.0,
+                        prio: 2,
+                    }),
+                    unary_op: Some(|_| 0.0),
+                };
+                DefaultOpsFactory::<f32>::make()
+                    .iter()
+                    .cloned()
+                    .chain(once(zero_mapper))
+                    .collect::<Vec<_>>()
+            }
+        }
+        let expr =
+            FlatEx::<f32, ExtendedF32Operators>::from_str("2^2*1/(berti) + zer0(4)").unwrap();
         let val = expr.eval(&[4.0]).unwrap();
         assert_float_eq_f32(val, 1.0);
     }
@@ -655,7 +669,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |_: f64| 1.0;
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -663,7 +677,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |_: f64| 1.0;
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -671,7 +685,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |_: f64| -1.0;
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -679,7 +693,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |_: f64| -1.0;
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -687,7 +701,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |_: f64| 1.0;
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -695,7 +709,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |x: f64| x.sin().cos() * x.cos();
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -703,9 +717,15 @@ mod tests {
         println!("{}", sut);
         let var_idx = 1;
         let n_vars = 2;
-        let flatex_1 = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |x: f64| x.cos() + x.sin();
-        test(var_idx, n_vars, -10000.0..10000.0, flatex_1.clone(), reference);
+        test(
+            var_idx,
+            n_vars,
+            -10000.0..10000.0,
+            flatex_1.clone(),
+            reference,
+        );
         let deri = flatex_1.partial(var_idx).unwrap();
         let reference = |x: f64| -x.sin() + x.cos();
         test(var_idx, n_vars, -10000.0..10000.0, deri.clone(), reference);
@@ -720,7 +740,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 1;
         let n_vars = 2;
-        let flatex_1 = parse_with_default_ops::<f64>("sin(x)-cos(x)+tan(x)+a").unwrap();
+        let flatex_1 = FlatEx::<f64>::from_str("sin(x)-cos(x)+tan(x)+a").unwrap();
         let reference = |x: f64| x.cos() + x.sin() + 1.0 / (x.cos().powf(2.0));
         test(var_idx, n_vars, -10000.0..10000.0, flatex_1, reference);
 
@@ -728,7 +748,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 1;
         let n_vars = 3;
-        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |x: f64| 1.0 / x * x.exp() + x.ln() * x.exp();
         test(var_idx, n_vars, 0.01..100.0, flatex, reference);
 
@@ -736,7 +756,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 2;
         let n_vars = 4;
-        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |x: f64| {
             (x.cosh() * x.cosh() - x.sinh() * x.sinh()) / x.cosh().powf(2.0)
                 + 1.0 / (x.cosh().powf(2.0))
@@ -747,7 +767,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 1;
         let n_vars = 4;
-        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex = FlatEx::<f64>::from_str(sut).unwrap();
         let reference = |x: f64| {
             1.0 / (1.0 - x.powf(2.0)).sqrt() - 1.0 / (1.0 - x.powf(2.0)).sqrt()
                 + 1.0 / (1.0 + x.powf(2.0))
@@ -758,7 +778,7 @@ mod tests {
         println!("{}", sut);
         let var_idx = 0;
         let n_vars = 1;
-        let flatex = parse_with_default_ops::<f64>(sut).unwrap();
+        let flatex = FlatEx::<f64>::from_str(sut).unwrap();
         let reference =
             |x: f64| 1.0 / (2.0 * x.sqrt()) * x.powf(1.57) + x.sqrt() * 1.57 * x.powf(0.57);
         test(var_idx, n_vars, 0.0..100.0, flatex, reference);
@@ -845,17 +865,17 @@ mod tests {
 
     #[test]
     fn test_error_handling() {
-        assert!(eval_str("").is_err());
-        assert!(eval_str("5+5-(").is_err());
-        assert!(eval_str(")2*(5+5)*3-2)*2").is_err());
-        assert!(eval_str("2*(5+5))").is_err());
+        assert!(eval_str::<f64>("").is_err());
+        assert!(eval_str::<f64>("5+5-(").is_err());
+        assert!(eval_str::<f64>(")2*(5+5)*3-2)*2").is_err());
+        assert!(eval_str::<f64>("2*(5+5))").is_err());
     }
 
-    #[cfg(feature = "serde_support")]
+    #[cfg(feature = "serde")]
     #[test]
     fn test_serde_public_interface() {
         let s = "{x}^(3.0-{y})";
-        let flatex = parse_with_default_ops::<f64>(s).unwrap();
+        let flatex = FlatEx::<f64>::from_str(s).unwrap();
         let serialized = serde_json::to_string(&flatex).unwrap();
         let deserialized = serde_json::from_str::<FlatEx<f64>>(serialized.as_str()).unwrap();
         assert_eq!(s, format!("{}", deserialized));
