@@ -1,24 +1,11 @@
-use crate::operators::Operator;
+use crate::definitions::N_NODES_ON_STACK;
+use crate::{ExError, ExResult, {operators::Operator}};
 use lazy_static::lazy_static;
 use regex::Regex;
 use smallvec::SmallVec;
-use std::error::Error;
-use std::fmt::{self, Debug};
+use std::fmt::Debug;
 use std::str::FromStr;
 
-/// This will be thrown at you if the parsing went wrong. Ok, obviously it is not an
-/// exception, so thrown needs to be understood figuratively.
-#[derive(Debug, Clone)]
-pub struct ExParseError {
-    pub msg: String,
-}
-impl fmt::Display for ExParseError {
-    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
-        write!(f, "{}", self.msg)
-    }
-}
-impl Error for ExParseError {}
-
 #[derive(Debug, PartialEq, Eq)]
 pub enum Paren {
     Open,
@@ -60,7 +47,6 @@ pub fn is_numeric_regex<'a>(re: &Regex, text: &'a str) -> Option<&'a str> {
     }
 }
 
-
 /// Parses tokens of a text with regexes and returns them as a vector
 ///
 /// # Arguments
@@ -73,22 +59,28 @@ pub fn is_numeric_regex<'a>(re: &Regex, text: &'a str) -> Option<&'a str> {
 ///
 /// See [`parse_with_number_pattern`](parse_with_number_pattern)
 ///
-pub fn tokenize_and_analyze<'a, T: Copy + FromStr + Debug, F: Fn(&'a str) -> Option<&'a str>>(
+pub fn tokenize_and_analyze<
+    'a,
+    T: Copy + FromStr + Debug,
+    F: Fn(&'a str) -> Option<&'a str>,
+>(
     text: &'a str,
     ops_in: &[Operator<'a, T>],
     is_numeric: F,
-) -> Result<Vec<ParsedToken<'a, T>>, ExParseError>
+) -> ExResult<SmallVec<[ParsedToken<'a, T>; N_NODES_ON_STACK]>>
 where
     <T as std::str::FromStr>::Err: Debug,
 {
     // Make sure that the text does not contain unicode characters
     if text.chars().any(|c| !c.is_ascii()) {
-        return Err(ExParseError{msg: "only ascii characters are supported".to_string()});
+        return Err(ExError {
+            msg: "only ascii characters are supported".to_string(),
+        });
     };
 
     // We sort operators inverse alphabetically such that log2 has higher priority than log (wlog :D).
     let mut ops_tmp = ops_in.iter().clone().collect::<SmallVec<[_; 64]>>();
-    ops_tmp.sort_unstable_by(|o1, o2| o2.repr.partial_cmp(o1.repr).unwrap());
+    ops_tmp.sort_unstable_by(|o1, o2| o2.repr().partial_cmp(o1.repr()).unwrap());
     let ops = ops_tmp; // from now on const
 
     lazy_static! {
@@ -97,15 +89,15 @@ where
 
     let find_ops = |offset: usize| {
         ops.iter().find(|op| {
-            let range_end = offset + op.repr.chars().count();
+            let range_end = offset + op.repr().chars().count();
             if range_end > text.len() {
                 false
             } else {
-                op.repr == &text[offset..range_end]
+                op.repr() == &text[offset..range_end]
             }
         })
     };
-    let mut res = Vec::new();
+    let mut res: SmallVec<[_; N_NODES_ON_STACK]> = SmallVec::new();
     let mut cur_offset = 0usize;
     for (i, c) in text.chars().enumerate() {
         if c == ' ' {
@@ -130,7 +122,7 @@ where
                 cur_offset += n_chars;
                 ParsedToken::<T>::Num(num_str.parse::<T>().unwrap())
             } else if let Some(op) = find_ops(cur_offset) {
-                let n_chars = op.repr.chars().count();
+                let n_chars = op.repr().chars().count();
                 cur_offset += n_chars;
                 ParsedToken::<T>::Op(**op)
             } else if let Some(var_str) = RE_NAME.find(text_rest) {
@@ -140,12 +132,11 @@ where
                 ParsedToken::<T>::Var(var_str)
             } else {
                 let msg = format!("how to parse the beginning of {}", text_rest);
-                return Err(ExParseError { msg });
+                return Err(ExError { msg });
             };
             res.push(next_parsed_token);
         }
     }
-    check_parsed_token_preconditions(&res)?;
     Ok(res)
 }
 
@@ -154,7 +145,8 @@ struct PairPreCondition<'a, 'b, T: Copy + FromStr> {
     error_msg: &'b str,
 }
 
-fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a, 'b, T>; 9] {
+fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>(
+) -> [PairPreCondition<'a, 'b, T>; 9] {
     [
         PairPreCondition {
             apply: |left, right| match (left, right) {
@@ -180,7 +172,8 @@ fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a
             apply: |left, right| match (left, right) {
                 (ParsedToken::Num(_), ParsedToken::Op(op))
                 | (ParsedToken::Var(_), ParsedToken::Op(op))
-                    if op.bin_op.is_none() =>
+                    // we do not ask for is_unary since operators can be both
+                    if !op.has_bin() =>
                 {
                     false
                 }
@@ -191,7 +184,7 @@ fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a
         PairPreCondition {
             apply: |left, right| match (left, right) {
                 (ParsedToken::Op(op_l), ParsedToken::Op(op_r))
-                    if op_l.unary_op.is_none() && op_r.unary_op.is_none() =>
+                    if !op_l.has_unary() && !op_r.has_unary() =>
                 {
                     false
                 }
@@ -202,7 +195,7 @@ fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a
         PairPreCondition {
             apply: |left, right| match (left, right) {
                 (ParsedToken::Op(op_l), ParsedToken::Op(op_r))
-                    if op_l.bin_op.is_none() && op_r.unary_op.is_none() =>
+                    if !op_l.has_bin() && !op_r.has_unary() =>
                 {
                     false
                 }
@@ -219,9 +212,7 @@ fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a
         },
         PairPreCondition {
             apply: |left, right| match (left, right) {
-                (ParsedToken::Paren(_p @ Paren::Close), ParsedToken::Op(op))
-                    if op.bin_op.is_none() =>
-                {
+                (ParsedToken::Paren(_p @ Paren::Close), ParsedToken::Op(op)) if !op.has_bin() => {
                     false
                 }
                 _ => true,
@@ -230,9 +221,7 @@ fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a
         },
         PairPreCondition {
             apply: |left, right| match (left, right) {
-                (ParsedToken::Paren(_p @ Paren::Open), ParsedToken::Op(op))
-                    if op.unary_op.is_none() =>
-                {
+                (ParsedToken::Paren(_p @ Paren::Open), ParsedToken::Op(op)) if !op.has_unary() => {
                     false
                 }
                 _ => true,
@@ -262,37 +251,39 @@ fn make_pair_pre_conditions<'a, 'b, T: Copy + FromStr>() -> [PairPreCondition<'a
 ///
 /// See [`parse_with_number_pattern`](parse_with_number_pattern)
 ///
-pub fn check_parsed_token_preconditions<T>(parsed_tokens: &[ParsedToken<T>]) -> Result<u8, ExParseError>
+pub fn check_parsed_token_preconditions<'a, T>(
+    parsed_tokens: &[ParsedToken<'a, T>],
+) -> ExResult<()>
 where
-    T: Copy + FromStr + std::fmt::Debug,
+    T: Copy + FromStr + Debug
 {
     if parsed_tokens.len() == 0 {
-        return Err(ExParseError {
+        return Err(ExError {
             msg: "cannot parse empty string".to_string(),
         });
     };
 
     let pair_pre_conditions = make_pair_pre_conditions::<T>();
     (0..parsed_tokens.len() - 1)
-        .map(|i| -> Result<(), ExParseError> {
+        .map(|i| {
             let failed = pair_pre_conditions
                 .iter()
                 .map(|ppc| (ppc, (ppc.apply)(&parsed_tokens[i], &parsed_tokens[i + 1])))
-                .find(|(_, ppc_passed)| !ppc_passed);
+                .find(|(_, ppc_passed)| !*ppc_passed);
             match failed {
-                Some((failed_ppc, _)) => Err(ExParseError {
+                Some((failed_ppc, _)) => Err(ExError {
                     msg: failed_ppc.error_msg.to_string(),
                 }),
                 None => Ok(()),
             }
         })
-        .collect::<Result<Vec<_>, _>>()?;
+        .collect::<ExResult<Vec<_>>>()?;
 
     let mut open_paren_cnt = 0i32;
     parsed_tokens
         .iter()
         .enumerate()
-        .map(|(i, expr_elt)| -> Result<(), ExParseError> {
+        .map(|(i, expr_elt)| -> ExResult<()> {
             match expr_elt {
                 ParsedToken::Paren(p) => {
                     open_paren_cnt += match p {
@@ -300,7 +291,7 @@ where
                         Paren::Open => 1,
                     };
                     if open_paren_cnt < 0 {
-                        return Err(ExParseError {
+                        return Err(ExError {
                             msg: format!("too many closing parentheses until position {}", i),
                         });
                     }
@@ -309,35 +300,22 @@ where
                 _ => Ok(()),
             }
         })
-        .collect::<Result<Vec<_>, _>>()?;
+        .collect::<ExResult<Vec<_>>>()?;
     if open_paren_cnt != 0 {
-        Err(ExParseError {
+        Err(ExError {
             msg: "parentheses mismatch".to_string(),
         })
     } else if let ParsedToken::Op(_) = parsed_tokens[parsed_tokens.len() - 1] {
-        Err(ExParseError {
+        Err(ExError {
             msg: "the last element cannot be an operator".to_string(),
         })
     } else {
-        Ok(0)
+        Ok(())
     }
 }
 
 #[cfg(test)]
-use crate::operators;
-#[test]
-fn test_tokenize_and_analze() {
-    let ops = operators::make_default_operators::<f32>();
-    
-    let text = r"5\6";
-    let elts = tokenize_and_analyze(text, &ops, is_numeric_text);
-    assert!(elts.is_err());
-
-    let text = "ӭ";
-    let elts = tokenize_and_analyze(text, &ops, is_numeric_text);
-    assert!(elts.is_err());
-
-}
+use crate::operators::{DefaultOpsFactory, MakeOperators};
 
 #[test]
 fn test_is_numeric() {
@@ -353,7 +331,7 @@ fn test_is_numeric() {
 #[test]
 fn test_preconditions() {
     fn test(text: &str, msg_part: &str) {
-        fn check_err_msg<V>(err: Result<V, ExParseError>, msg_part: &str) {
+        fn check_err_msg<X>(err: ExResult<X>, msg_part: &str) {
             match err {
                 Ok(_) => {
                     println!("expected an error that should contain '{}'", msg_part);
@@ -366,18 +344,18 @@ fn test_preconditions() {
                 }
             }
         }
-        let ops = operators::make_default_operators::<f32>();
+        let ops = DefaultOpsFactory::<f32>::make();
         let elts = tokenize_and_analyze(text, &ops, is_numeric_text);
         match elts {
-            Ok(elts_unwr) => {
-                let err = check_parsed_token_preconditions(&elts_unwr[..]);
-                check_err_msg(err, msg_part);
+            Err(e) => check_err_msg::<Vec<ParsedToken<f32>>>(Err(e), msg_part),
+            Ok(elts) => {
+                let error = check_parsed_token_preconditions(&elts);
+                check_err_msg(error, msg_part);
             }
-            Err(_) => check_err_msg(elts, msg_part),
-        }
+        };
     }
-    test("xo-17-(((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((expWW-tr-3746-4+sinnex-nn--nnexpWW-tr-7492-4+4-nsqrnexq+---------282)-384", "parentheses mismatch");
-    test("fi.g", "parse the beginning of .g");
+    // test("xo-17-(((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((((expWW-tr-3746-4+sinnex-nn--nnexpWW-tr-7492-4+4-nsqrnexq+---------282)-384", "parentheses mismatch");
+    // test("fi.g", "parse the beginning of .g");
     test("(nc7)sqrtE", "unary operator cannot be on the right");
     test("", "empty string");
     test("++", "the last element cannot be an operator");
